from django.urls import path

from simple_locations.api import api
from simple_locations.models import Area
from simple_locations.views import (
    AreaJSONLayerView,
    ChildAreasJSONLayerView,
    add_location,
    area_search,
    delete_location,
    edit_location,
    render_location,
    simple_locations,
)

app_name = "simple_locations"

urlpatterns = [
    path("simple_locations/edit/<int:area_id>/", edit_location),
    path("simple_locations/add/", add_location),
    path("simple_locations/add/<int:parent_id>/", add_location),
    path("simple_locations/delete/<int:area_id>/", delete_location),
    path("simple_locations/render_tree/", render_location),
    path("simple_locations/", simple_locations),
    path("areasearch/", area_search),
    path(
        "data.geojson/",
        AreaJSONLayerView.as_view(model=Area, properties=("name", "area_activities__activity__name")),
        name="data",
    ),
    path(
        "<int:area>/children.geojson/",
        ChildAreasJSONLayerView.as_view(model=Area, properties=("name", "id", "code", "parent_id")),
        name="children",
    ),
    path("api/", api.urls),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def load_initial_data(apps, schema_editor):
    # this migration used to load the initial data, we do not want this anymore, but are leaving it here
    # to preserve migration history
    # call_command('loaddata', 'initial_data', app_label='simple_locations')
    pass


class Migration(migrations.Migration):

    dependencies = [("simple_locations", "0001_initial")]

    operations = [
        migrations.RunPython(load_initial_data),
    ]



from typing import Generator, List, Tuple

from ninja import Router

from simple_locations import model_schemas, models, schemas

router = Router(tags=["SimpleLocations"])

SIMPLIFICATION_LEVELS: Tuple[float, ...] = (0.0, 0.0001, 0.001, 0.01, 0.1, 0.5)


@router.get("/area/list.json", response=List[model_schemas.AreaModelSchema])
def area_list(request):
    """
    Returns a list of area information: name, id, parent id
    """
    return models.Area.objects.all()


@router.get("/areatype/list.json", response=List[model_schemas.AreaTypeModelSchema])
def area_type_list(request):
    """
    List the different area types
    """
    return models.AreaType.objects.all()


@router.get("/area/by-id/{area_id}.geojson", response=schemas.Feature)
def area_id(request, area_id: int):
    """
    Returns the geometry of a single Area
    as a single GeoJSON Feature
    """
    features: Generator[schemas.Feature, None, None] = models.Area.features.filter(pk=area_id).to_features()
    return schemas.Feature.parse_obj(next(features))


@router.get("/area/by-parent/{area_id}-s{simplify}-q{quantize}.geojson", response=schemas.FeatureCollection)
def area_children_compressed(request, area_id: int, simplify: int, quantize: int):
    """
    Returns the direct descendants of a given Area as a FeatureCollection
    applying compression methods
    """
    return models.Area.features.filter(parent=area_id).to_featurecollection(
        simplify=SIMPLIFICATION_LEVELS[simplify], quantize=quantize
    )


@router.get("/area/by-parent/{area_id}.geojson", response=schemas.FeatureCollection)
def area_children(request, area_id: int):
    """
    Returns the direct descendants of a given Area as a FeatureCollection
    applying compression methods
    """
    return models.Area.features.filter(parent=area_id).to_featurecollection()


@router.get("/area/by-type/{area_type}-s{simplify}-q{quantize}.geojson", response=schemas.FeatureCollection)
def area_type_compressed(request, area_type: str, simplify: int, quantize: int):
    """
    Returns all areas of a given type
    appliying simplification and quantization
    """
    return models.Area.features.filter(kind__slug=area_type).to_featurecollection(
        simplify=SIMPLIFICATION_LEVELS[simplify], quantize=quantize
    )


@router.get("/area/by-type/{area_type}.geojson", response=schemas.FeatureCollection)
def area_type(request, area_type: str):
    """
    Returns all areas of a given type
    """
    return models.Area.features.filter(kind__slug=area_type).to_featurecollection()



from django.core.management.base import BaseCommand

from simple_locations.models import Area


class Command(BaseCommand):
    help = """prints all districts in a form that's easy to import into excel"""

    def handle(self, **options):
        print("id\tname\ttype\tparent_id\tparent_name\tparent_type")
        for a in Area.objects.all():
            print(
                "%d\t%s\t%s\t%d\t%s\t%s"
                % (
                    a.pk,
                    a.name,
                    a.kind.name,
                    a.parent.pk if a.parent else -1,
                    a.parent.name if a.parent else "",
                    a.parent.kind.name if a.parent else "",
                )
            )



from typing import Generator, Union

from django.db import models

from simple_locations.gis_functions import JsonFeature
from simple_locations.schemas import Feature, FeatureCollection


class FeatureQueryset(models.QuerySet):
    def to_featurecollection(
        self, simplify: Union[float, None] = None, quantize: Union[int, None] = None
    ) -> FeatureCollection:
        """
        GeoJSON Feature builder
        =======================

        >>> queryset = Area.geofunctions.all()
        >>> # Annotate a "feature" field on to the queryset
        >>> queryset.annotate(JsonFeature())
        >>> # Optional simplify, quantize the results:
        >>> queryset.annotate(JsonFeature(simplify=1e-3, quantize=5))
        >>> # Return an array for a FeatureCollection instance
        >>> queryset.aggregate(features = JSONBAgg(JsonFeature()))

        Pydantic integration
        ====================
        >>> from simple_locations.schemas import Feature
        >>> Feature.parse_obj(Area.geofunctions.all().annotate(JsonFeature()).first().feature)
        >>> # A list of features:
        >>> [Feature.parse_obj(i) for i in Area.geofunctions.filter(kind__name='district').to_features(simplify=1e-3, quantize=5).values_list('feature', flat=True)]
        >>> # A FeatureCollection:
        >>> FeatureCollection.parse_obj(queryset.aggregate(features = JSONBAgg(JsonFeature())))
        """
        return FeatureCollection.construct(features=[*self.to_features(simplify=simplify, quantize=quantize)])

    def annotate_features(self, simplify: Union[float, None] = None, quantize: Union[int, None] = None):
        """
        Annotated 'feature' fields onto the queryset
        """
        return self.annotate(
            JsonFeature(
                simplify=simplify,
                quantize=quantize,
                # The following fields become "properties"
                id=models.F("id"),
                parent=models.F("parent"),
                code=models.F("code"),
                name=models.F("name"),
                kind=models.F("kind__name"),
            )
        )

    def to_features(
        self, simplify: Union[float, None] = None, quantize: Union[int, None] = None
    ) -> Generator[Feature, None, None]:
        """
        Generate features
        """
        return (Feature(**area.feature) for area in self.annotate_features(simplify=simplify, quantize=quantize))


class FeatureManager(models.Manager):
    def get_queryset(self):
        return FeatureQueryset(self.model, using=self._db)

    def to_features(self, simplify: Union[float, None] = None, quantize: Union[int, None] = None):
        return self.get_queryset().to_features(simplify=simplify, quantize=quantize)

    def to_featurecollection(
        self, simplify: Union[float, None] = None, quantize: Union[int, None] = None
    ) -> FeatureCollection:
        return self.get_queryset().to_featurecollection(simplify=simplify, quantize=quantize)



from django.core.management.base import BaseCommand
from django.db import connection


class Scripts:
    init_topology = """
        -- Note:

        -- You may need to run the following SQL:

        -- CREATE EXTENSION postgis_topology;
        -- GRANT ALL ON SCHEMA topology TO dird;
        -- GRANT ALL ON ALL FUNCTIONS IN SCHEMA topology TO dird;
        -- GRANT ALL ON ALL TABLES IN SCHEMA topology TO dird;
        -- GRANT ALL ON ALL SEQUENCES IN SCHEMA topology TO dird;

        -- before running this script

        SELECT topology.CreateTopology('ma_topo', 3857, 15);
        CREATE TABLE if not exists simple_locations_topology(id serial, area_id int);
        SELECT topology.AddTopoGeometryColumn('ma_topo', 'public', 'simple_locations_topology', 'topo', 'POLYGON');
    """

    insert_areas = """
        -- This is fast
        insert into simple_locations_topology(area_id, topo)
            select id, topology.toTopoGeom(st_transform(geom, 3857), 'ma_topo', 1)
            from simple_locations_area where kind_id = 4;

    """

    insert_areas_2 = """
        -- This took about 3 - 4 minutes
        -- Topology inserts of polygons are known to be very slow
        insert into simple_locations_topology(area_id, topo)
            select id, topology.toTopoGeom(st_transform(geom, 3857), 'ma_topo', 1)
            from simple_locations_area where kind_id != 4;
    """

    populate_edge_table = """
        truncate simple_locations_border cascade;
        insert into simple_locations_border(id, geom, area_ids, area_types)
            select distinct e.edge_id, e.geom, ARRAY[]::int[], ARRAY[]::int[] from
            ma_topo.edge e;
    """

    populate_border_areas = """
        -- Populate the border -> area relations table
        truncate simple_locations_border_area CASCADE;

        -- "Outer" lines are where a line does not occur more than once
        -- for faces within a polygon

        -- The faces for a polygon are a relation between
        -- the toplogy layer to face:
        -- [simple_locations_topology] --> [topo.relation] --> [topo.face]

        with relation_faces as (
            -- Join faces table to relations table where the
            -- relation is valid. This should be all rows
            -- but just to be sure filter to type=3 and layer_id = 1
            select face_id, topogeo_id from ma_topo.face, ma_topo.relation r2
            where r2.element_type = 3  -- 1=point, 2=line, 3=polygon
            and r2.element_id = face.face_id
            and layer_id = 1
        )
        , area_faces as (
            -- Join from the topology to faces
            -- The resuult of this would be to match an 'area' to topo 'faces'
            -- Generally 'faces' here would be the smallest unit added (LLG level for PNG)
            select face_id, area_id from simple_locations_topology slt ,
            relation_faces rf
            where (slt.topo).id = rf.topogeo_id
        )

        , area_to_edge as (
            select ed.edge_id, area_faces.area_id from ma_topo.edge_data ed ,
            area_faces
            where ed.left_face = area_faces.face_id
            or ed.right_face = area_faces.face_id
        )
        -- Here a good troubleshooting return would be:
        -- select st_collect(geom) from area_to_edge natural left join ma_topo.edge_data where area_id = 822
        -- This will include all interior and exterior lines
        -- In order to correct this we need to drop any lines which are not part
        -- of only one face...
        --select * from area_to_edge natural left join ma_topo.edge_data where area_id = 822

        , count_edges as (
            select edge_id, count(edge_id), area_id from area_to_edge group by edge_id, area_id
        )

        --select st_collect(edge_data.geom) from count_edges, ma_topo.edge_data
        --	where count_edges.edge_id = edge_data.edge_id
        --	and count = 1
        --	and area_id = 822

        insert into simple_locations_border_area (border_id, area_id)
        select edge_id, area_id from count_edges where count = 1;
    """

    update_border_area_array = """
        with areas as (select border_id, array_agg(area_id) as area_ids from simple_locations_border_area group by border_id)
        update simple_locations_border b set area_ids =
            (select areas.area_ids from areas
            where areas.border_id = b.id)
            where b.id in (select border_id from areas)
    """

    update_border_type_array = """

        with area_types as (
            select border_id, array_agg(distinct slt.id) area_types from simple_locations_area sla,
                simple_locations_areatype slt,
                simple_locations_border_area slba
                where sla.kind_id = slt.id
                and slba.area_id = sla.id
                group by border_id
            ) update simple_locations_border b
            set area_types = (
                select area_types from area_types
                where area_types.border_id = b.id)
                where b.id in (select border_id from area_types)
    """

    remove_topology = """
        -- Undo all the changes, remove the topology layer
        select topology.DropTopoGeometryColumn('public', 'simple_locations_topology', 'topo');
        drop table simple_locations_topology;
        select topology.DropTopology('ma_topo')
    """


class Command(BaseCommand):
    help = "My shiny new management command."

    def add_arguments(self, parser):
        ...
        # parser.add_argument('sample', nargs='+')

    def handle(self, *args, **options):
        with connection.cursor() as c:
            self.stdout.write(self.style.SUCCESS(Scripts.init_topology))
            c.execute(Scripts.init_topology)
            self.stdout.write(self.style.SUCCESS(Scripts.insert_areas))
            c.execute(Scripts.insert_areas)
            self.stdout.write(self.style.SUCCESS(Scripts.insert_areas_2))
            c.execute(Scripts.insert_areas_2)
            self.stdout.write(self.style.SUCCESS(Scripts.populate_edge_table))
            c.execute(Scripts.populate_edge_table)
            self.stdout.write(self.style.SUCCESS(Scripts.populate_border_areas))
            c.execute(Scripts.populate_border_areas)

            self.stdout.write(self.style.SUCCESS(Scripts.update_border_type_array))
            c.execute(Scripts.update_border_type_array)

            self.stdout.write(self.style.SUCCESS(Scripts.update_border_area_array))
            c.execute(Scripts.update_border_area_array)

            self.stdout.write(self.style.SUCCESS(Scripts.remove_topology))
            c.execute(Scripts.remove_topology)

        # raise NotImplementedError()



# Generated by Django 3.1.13 on 2022-02-21 08:17

import django.contrib.gis.db.models.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('simple_locations', '0011_denormed_border_fields'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectedArea',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('geom', django.contrib.gis.db.models.fields.MultiPolygonField(blank=True, null=True, srid=3857)),
                ('area', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='simple_locations.area')),
            ],
        ),
        migrations.RunSQL(
            "TRUNCATE simple_locations_projectedarea; INSERT INTO simple_locations_projectedarea (geom, area_id) (SELECT st_transform(geom, 3857), id FROM simple_locations_area);",
            reverse_sql=None
        )
    ]



from django.conf import settings
from django.db import IntegrityError
from django.db.models import Q
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.template import RequestContext
from django.views.decorators.cache import cache_control
from djgeojson.views import GeoJSONLayerView
from mptt.exceptions import InvalidMove

from simple_locations.models import Area, AreaType, Point

from .forms import LocationForm


@cache_control(no_cache=True)
def simple_locations(request):
    """
    firefox likes to aggressively cache forms set cache control to false to override this
    """
    form = LocationForm()
    nodes = Area.objects.all()
    return render(
        request,
        "simple_locations/index.html",
        {
            "form": form,
            "nodes": nodes,
            "map_key": settings.MAP_KEY,
        },
        context_instance=RequestContext(request),
    )


def add_location(req, parent_id=None):
    nodes = Area.objects.all()

    if req.method == "POST":
        form = LocationForm(req.POST)
        if form.is_valid():
            name = form.cleaned_data["name"]
            code = form.cleaned_data["code"]
            lat = form.cleaned_data["lat"]
            lon = form.cleaned_data["lon"]
            target = form.cleaned_data["target"]
            kind = form.cleaned_data["kind"]
            area = Area.objects.create(name=name, code=code, parent=target)
            if lat and lon:
                location = Point(latitude=lat, longitude=lon)
                location.save()
                area.location = location
            try:
                kind = get_object_or_404(AreaType, pk=int(kind))
                area.kind = kind
            except ValueError:
                pass
            area.save()
            form = LocationForm()

            return render(
                req,
                "simple_locations/location_edit.html",
                {"form": form, "nodes": nodes},
                context_instance=RequestContext(req),
            )
        else:
            form = LocationForm(req.POST)
            return render(
                req,
                "simple_locations/location_edit.html",
                {"form": form, "nodes": nodes},
                context_instance=RequestContext(req),
            )

    else:
        if parent_id:
            default_data = {}
            parent = get_object_or_404(Area, pk=parent_id)
            default_data["move_choice"] = True
            default_data["target"] = parent.pk
            default_data["position"] = "last-child"
            form = LocationForm(default_data)
            form._errors = ""
        else:
            form = LocationForm()

    return render(
        req,
        "simple_locations/location_edit.html",
        {"form": form, "nodes": nodes},
        context_instance=RequestContext(req),
    )


def edit_location(req, area_id):
    location = get_object_or_404(Area, pk=area_id)
    if req.method == "POST":
        form = LocationForm(req.POST)
        if form.is_valid():
            saved = True

            area = Area.objects.get(pk=area_id)
            area.name = form.cleaned_data["name"]
            area.code = form.cleaned_data["code"]
            lat = form.cleaned_data["lat"]
            lon = form.cleaned_data["lon"]
            kind = form.cleaned_data["kind"]

            try:
                kind = get_object_or_404(AreaType, pk=int(kind))
                area.kind = kind
            except ValueError:
                pass
            if lat and lon:
                point = Point(latitude=lat, longitude=lon)
                point.save()
                area.location = point
            try:
                area.save()
            except IntegrityError:
                form.errors["code"] = "This code already exists"
                saved = False

            if form.cleaned_data["move_choice"]:
                target = form.cleaned_data["target"]

                try:
                    area.parent = target
                    area.save()
                except InvalidMove:
                    form.errors["position"] = "This move is invalid"
                    saved = False

            if saved:
                form = LocationForm()
                return render(
                    req,
                    "simple_locations/location_edit.html",
                    {"form": form, "nodes": Area.objects.all()},
                    context_instance=RequestContext(req),
                )
            else:
                return render(
                    req,
                    "simple_locations/location_edit.html",
                    {"form": form, "item": location, "nodes": Area.objects.all()},
                    context_instance=RequestContext(req),
                )

        else:
            return render(
                req,
                "simple_locations/location_edit.html",
                {"form": form, "item": location},
                context_instance=RequestContext(req),
            )
    else:
        default_data = {}
        default_data["pk"] = location.pk
        default_data["name"] = location.name
        default_data["code"] = location.code
        default_data["move_choice"] = False
        if location.kind:
            default_data["kind"] = location.kind.pk
        if location.parent:
            default_data["target"] = location.parent
            default_data["position"] = "last-child"
        if location.location:
            default_data["lat"] = location.location.latitude
            default_data["lon"] = location.location.longitude
        form = LocationForm(default_data)
        return render(
            req,
            "simple_locations/location_edit.html",
            {"form": form, "nodes": Area.objects.all(), "item": location},
            context_instance=RequestContext(req),
        )


def delete_location(request, area_id):
    node = get_object_or_404(Area, pk=area_id)
    if request.method == "POST":
        node.delete()

    return HttpResponseRedirect("/simple_locations/render_tree")


@cache_control(no_cache=True)
def render_location(request):
    nodes = Area.objects.all()
    return render(request, "simple_locations/treepanel.html", {"nodes": nodes})


def area_search(request):
    import json

    areadetails = []
    if request.GET.__contains__("query"):
        objects = Area.objects.filter(name__iregex=request.GET["query"])
    else:
        objects = Area.objects.all()
    for area in objects:
        areadetail = {}
        areadetail["value"] = area.pk
        areadetail["name"] = area.name
        areadetail["kind"] = area.kind.name

        if area.parent:
            areadetail["parentname"] = area.parent.name
            areadetail["parentkind"] = area.parent.kind.name
        areadetails.append(areadetail)
    return HttpResponse(json.dumps(areadetails))


class AreaJSONLayerView(GeoJSONLayerView):
    precision = 3
    simplify = 0.002

    def get_queryset(self, *args, **kwargs):
        areas = [i for i in self.request.GET.getlist("locations[]", []) if i.isnumeric()]
        queryset = (
            super(AreaJSONLayerView, self).get_queryset(*args, **kwargs).filter(Q(geom__isnull=False, pk__in=areas))
        )
        return queryset


class ChildAreasJSONLayerView(GeoJSONLayerView):
    precision = 3
    simplify = 0.002

    def get_queryset(self, *args, **kwargs):
        area = self.kwargs.get("area", self.request.GET.get("area", 1))
        queryset = (
            super(ChildAreasJSONLayerView, self)
            .get_queryset(*args, **kwargs)
            .filter(Q(geom__isnull=False, parent_id=area))
        )
        return queryset



# Generated by Django 3.1rc1 on 2020-08-04 05:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('simple_locations', '0007_area_geom_can_be_blank'),
    ]

    operations = [
        migrations.CreateModel(
            name='AreaProfile',
            fields=[
                ('date_created', models.DateField(auto_now_add=True, null=True, verbose_name='Date Created')),
                ('date_modified', models.DateField(blank=True, null=True, verbose_name='Last Modified')),
                ('area', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='simple_locations.area')),
                ('description', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='IndicatorMeasureSchema',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('schema', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='AreaIndicator',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateField(auto_now_add=True, null=True, verbose_name='Date Created')),
                ('date_modified', models.DateField(blank=True, null=True, verbose_name='Last Modified')),
                ('name', models.TextField()),
                ('measure', models.CharField(choices=[('U', 'Units'), ('P', 'Percentages'), ('N', 'Nominal'), ('O', 'Ordinal'), ('Q', 'Qualitative')], default='U', help_text='Define the unit of measure in which the value is reported.', max_length=2)),
                ('value', models.JSONField()),
                ('dimensions', models.JSONField(help_text='A category used for disaggregating the result by gender, age, etc.')),
                ('area', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='simple_locations.area')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from typing import Any, Dict, List, Literal, Tuple, Union

from pydantic import BaseModel, Field

Coords = Tuple[float, float]
CoordsA = List[Coords]


class Point(BaseModel):
    type: Literal["Point"]
    coordinates: Coords


class MultiPoint(BaseModel):
    type: Literal["MultiPoint"]
    coordinates: CoordsA


class Polygon(BaseModel):
    type: Literal["Polygon"]
    coordinates: Union[List[Coords], List[CoordsA]]


class MultiPolygon(BaseModel):
    type: Literal["MultiPolygon"]
    coordinates: Union[
        List[List[Coords]],
        List[List[List[Coords]]],
        List[List[List[List[Coords]]]],
    ]


class Feature(BaseModel):
    type: Literal["Feature"] = "Feature"
    geometry: Union[Point, MultiPoint, Polygon, MultiPolygon] = Field(..., discriminator="type")
    id: Union[int, str]
    properties: Dict[str, Any]


class FeatureCollection(BaseModel):
    type: Literal["FeatureCollection"] = "FeatureCollection"
    features: List[Feature]



# Generated by Django 3.1.13 on 2022-02-17 11:02

import django.contrib.gis.db.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("simple_locations", "0009_auto_20211111_0805"),
    ]

    operations = [
        migrations.CreateModel(
            name="Border",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("geom", django.contrib.gis.db.models.fields.LineStringField(srid=3857)),
                ("area", models.ManyToManyField(to="simple_locations.Area")),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from typing import List, Tuple

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]  # type: List[Tuple[str, str]]

    operations = [
        migrations.CreateModel(
            name="Area",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                ("name", models.CharField(max_length=100)),
                ("code", models.CharField(max_length=50)),
                ("lft", models.PositiveIntegerField(editable=False, db_index=True)),
                ("rght", models.PositiveIntegerField(editable=False, db_index=True)),
                ("tree_id", models.PositiveIntegerField(editable=False, db_index=True)),
                ("level", models.PositiveIntegerField(editable=False, db_index=True)),
            ],
            options={
                "verbose_name": "Area",
                "verbose_name_plural": "Areas",
            },
        ),
        migrations.CreateModel(
            name="AreaType",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                ("name", models.CharField(max_length=100)),
                ("slug", models.CharField(unique=True, max_length=30)),
            ],
            options={
                "verbose_name": "Area Type",
                "verbose_name_plural": "Area Types",
            },
        ),
        migrations.CreateModel(
            name="Facility",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                ("name", models.CharField(max_length=100)),
                ("code", models.CharField(max_length=64, blank=True)),
                ("area", models.ForeignKey(related_name="facility", blank=True, to="simple_locations.Area", null=True, on_delete=models.CASCADE)),
                ("catchment_areas", models.ManyToManyField(related_name="catchment", to="simple_locations.Area")),
            ],
            options={
                "verbose_name": "Facility",
                "verbose_name_plural": "Facilities",
            },
        ),
        migrations.CreateModel(
            name="FacilityType",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                ("name", models.CharField(max_length=100)),
                ("slug", models.CharField(unique=True, max_length=30)),
            ],
            options={
                "verbose_name": "Facility Type",
                "verbose_name_plural": "Facility Types",
            },
        ),
        migrations.CreateModel(
            name="Point",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                ("latitude", models.DecimalField(max_digits=13, decimal_places=10)),
                ("longitude", models.DecimalField(max_digits=13, decimal_places=10)),
            ],
            options={
                "verbose_name": "Point",
                "verbose_name_plural": "Points",
            },
        ),
        migrations.AddField(
            model_name="facility",
            name="location",
            field=models.ForeignKey(blank=True, to="simple_locations.Point", null=True, on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name="facility",
            name="parent",
            field=models.ForeignKey(related_name="facility", blank=True, to="simple_locations.Facility", null=True, on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name="facility",
            name="type",
            field=models.ForeignKey(blank=True, to="simple_locations.FacilityType", null=True, on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name="area",
            name="kind",
            field=models.ForeignKey(blank=True, to="simple_locations.AreaType", null=True, on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name="area",
            name="location",
            field=models.ForeignKey(blank=True, to="simple_locations.Point", null=True, on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name="area",
            name="parent",
            field=models.ForeignKey(related_name="children", blank=True, to="simple_locations.Area", null=True, on_delete=models.CASCADE),
        ),
        migrations.AlterUniqueTogether(
            name="area",
            unique_together=set([("code", "kind")]),
        ),
    ]



from typing import List, Optional, TypeVar, Union

# We import this and then cast to JSON in the db. The default function returns text.
from django.contrib.gis.db.models.functions import AsGeoJSON as AsGeoJson_
from django.db import models
from django.db.models import F, Func, Value
from django.db.models.fields.json import JSONField
from django.db.models.functions.comparison import JSONObject


class Quantize(models.Func):
    """
    ST_QuantizeCoordinates determines the number of bits (N)
    required to represent a coordinate value with a specified number
    of digits after the decimal point, and then sets all but the N
    most significant bits to zero. The resulting coordinate value will
    still round to the original value, but will have improved compressiblity
    """

    function = "ST_QuantizeCoordinates"
    template = "%(function)s(%(expressions)s, %(quantize)s)"


class Simplify(models.Func):
    """
    Returns a "simplified" version of the given geometry using the Douglas-Peucker algorithm.
    """

    function = "ST_SIMPLIFY"
    template = "%(function)s(%(expressions)s, %(simplify)s)"


class SimplifyPreserve(Simplify):
    """
    Returns a "simplified" version of the given geometry using the Douglas-Peucker algorithm.
    """

    function = "ST_SIMPLIFYPRESERVETOPOLOGY"
    template = "%(function)s(%(expressions)s, %(simplify)s)"


class Multi(models.Func):
    """
    Returns the geometry as a MULTI* geometry collection.
    If the geometry is already a collection, it is returned unchanged.
    """

    function = "ST_Multi"


M = TypeVar("M", bound=models.Model)


class AsGeoJson(AsGeoJson_):
    """
    Official Django function returns text. This wraps that in a JSON cast
    and tells Python to expect a JSONField
    """

    ...
    template = f"{AsGeoJson_.template}::json"
    # output_field = models.JSONField()


class JsonFeature(models.Func):
    """
    Modern (Django 3.2+) GeoJSON object simplifier and serializer
    """

    function = "JSONB_BUILD_OBJECT"
    output_field = JSONField()
    default_alias = "feature"
    maxdecimaldigits = 6

    def __init__(
        self,
        # These are optional geometry processing functions which affect the
        # geometry before it's converted to GeoJSON
        geom_field: str = "geom",
        simplify: Optional[float] = None,
        quantize: Optional[int] = None,
        multi: Optional[bool] = False,
        # Include the Primary Key as ID for the model?
        # Primary key will be cast to an int or string, if possible
        include_id_field: Optional[bool] = True,
        # Additional fields for "ST_AsGeoJson"
        # include bbox, crs, precision
        # These are sufffixed with `_` to distinguish them from `fields`
        bbox_: bool = False,
        crs_: bool = False,
        precision_: int = 6,
        # Properties ('fields' or 'annotations' on the model)
        # which will become part of the GIS "properties" dict
        **fields,
    ):
        expressions: List[Union[Value, F, Func]] = []
        if include_id_field:
            expressions.extend((Value("id"), models.F("pk")))
        g: Union[F, Func] = models.F(geom_field)
        if simplify:
            g = SimplifyPreserve(g, simplify=simplify)
        if quantize:
            g = Quantize(g, quantize=quantize)
        if multi:
            g = Multi(g)
        expressions.extend((Value("geometry"), AsGeoJson(g, bbox=bbox_, crs=crs_, precision=precision_)))
        expressions.extend((Value("properties"), JSONObject(**fields)))
        super().__init__(*expressions)



from django.core.management.base import BaseCommand
from django.db import connection


class Scripts:
    populate_simple_locations_projectedarea = """
        TRUNCATE simple_locations_projectedarea;
        INSERT INTO simple_locations_projectedarea (geom, area_id)
        SELECT ST_TRANSFORM(geom, 3857), id FROM simple_locations_area;
    """


class Command(BaseCommand):
    help = "My shiny new management command."

    def add_arguments(self, parser):
        pass

    def handle(self, *args, **options):
        with connection.cursor() as c:
            self.stdout.write(self.style.SUCCESS(Scripts.populate_simple_locations_projectedarea))
            c.execute(Scripts.populate_simple_locations_projectedarea)



from ninja import ModelSchema

from simple_locations.models import Area, AreaType


class AreaModelSchema(ModelSchema):
    class Config:
        model = Area
        model_fields = ["name", "id", "kind", "parent"]


class AreaTypeModelSchema(ModelSchema):
    class Config:
        model = AreaType
        model_fields = ["id", "name", "slug"]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-03-30 07:18
from __future__ import unicode_literals

import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('simple_locations', '0003_auto_20170320_1024'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='facility',
            name='area',
        ),
        migrations.RemoveField(
            model_name='facility',
            name='catchment_areas',
        ),
        migrations.RemoveField(
            model_name='facility',
            name='location',
        ),
        migrations.RemoveField(
            model_name='facility',
            name='parent',
        ),
        migrations.RemoveField(
            model_name='facility',
            name='type',
        ),
        migrations.AlterModelManagers(
            name='area',
            managers=[
            ],
        ),
        migrations.AddField(
            model_name='area',
            name='geom',
            field=django.contrib.gis.db.models.fields.MultiPolygonField(null=True, srid=4326),
        ),
        migrations.DeleteModel(
            name='Facility',
        ),
        migrations.DeleteModel(
            name='FacilityType',
        ),
    ]



from django.conf import settings
from django.contrib.gis import admin
from mptt.admin import MPTTModelAdmin

from simple_locations.models import Area, AreaIndicator, AreaType, Point

area_admin_classes = [admin.OSMGeoAdmin, MPTTModelAdmin]

# Translated models are "nice to have"
try:
    from modeltranslation.admin import TranslationAdmin

    area_admin_classes.append(TranslationAdmin)
    area_type_admin = TranslationAdmin
except ImportError:
    TranslationAdmin = admin.ModelAdmin
    pass


class PointAdmin(admin.ModelAdmin):
    list_display = ("id", "latitude", "longitude")


class AreaTypeAdmin(TranslationAdmin):
    list_display = ("slug", "name")


class AreaChildrenInline(admin.TabularInline):
    model = Area
    fields = ["name", "code", "kind"]
    show_change_link = True
    extra = 0


area_admin_settings = {
    "default_lon": getattr(settings, "GIS_DEFAULT_LAT", -8.8742),
    "default_lat": getattr(settings, "GIS_DEFAULT_LON", 125.7275),
    "default_zoom": 16,
    "map_template": "simple_locations/admin/osm.html",
    # "debug": True  - enable if copy/pasting WKTs is useful,
    "units": "km",
    "list_display": ("name", "kind", "location", "code"),
    "search_fields": ["code", "name"],
    "list_filter": ("kind",),
    "related_search_fields": {"parent": ("^name",)},
    "inlines": [AreaChildrenInline],
}

AreaAdminClass = type("AreaAdminClass", tuple(area_admin_classes), area_admin_settings)

admin.site.register(Point, PointAdmin)
admin.site.register(AreaType, AreaTypeAdmin)
admin.site.register(Area, AreaAdminClass)
admin.site.register(AreaIndicator)



from modeltranslation.translator import TranslationOptions, translator

from .models import Area, AreaType


class AreaNameTranslations(TranslationOptions):
    fields = ("name",)


class AreaTypeTranslations(TranslationOptions):
    fields = ("name",)


translator.register(Area, AreaNameTranslations)
translator.register(AreaType, AreaTypeTranslations)
